"""
=========================================================================
test_cases.py
=========================================================================
Centralized test case repository for the generic backend.

Author : Peitian Pan
Date   : Dec 12, 2019
"""

from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.testcases import (
    CaseArrayBits32IfcInComp,
    CaseBits32ArrayClosureConstruct,
    CaseBits32ClosureConstruct,
    CaseBits32ConnectSubCompAttrComp,
    CaseBits32ConstBitsToTmpVarComp,
    CaseBits32ConstIntToTmpVarComp,
    CaseBits32FreeVarToTmpVarComp,
    CaseBits32IfcTmpVarOutComp,
    CaseBits32MultiTmpWireComp,
    CaseBits32PortOnly,
    CaseBits32TmpWireAliasComp,
    CaseBits32TmpWireComp,
    CaseBits32Wirex5DrivenComp,
    CaseBits32x5PortOnly,
    CaseBitSelOverBitSelComp,
    CaseBitSelOverPartSelComp,
    CaseComponentArgsComp,
    CaseComponentDefaultArgsComp,
    CaseConnectArrayNestedIfcComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectPortIndexComp,
    CaseConnectSliceToOutComp,
    CaseConnectSubCompIfcHierarchyComp,
    CaseConnectValRdyIfcComp,
    CaseDoubleStarArgComp,
    CaseMixedDefaultArgsComp,
    CaseNestedPackedArrayStructComp,
    CaseNestedStructPortOnly,
    CasePartSelOverBitSelComp,
    CasePartSelOverPartSelComp,
    CaseStarArgComp,
    CaseStructConstComp,
    CaseStructIfcTmpVarOutComp,
    CaseStructPortOnly,
    CaseStructTmpWireComp,
    CaseStructWireDrivenComp,
    CaseStructx5PortOnly,
    CaseSubCompFreeVarDrivenComp,
    CaseSubCompTmpDrivenComp,
    CaseTwoUpblksFreevarsComp,
    CaseTwoUpblksSliceComp,
    CaseTwoUpblksStructTmpWireComp,
    CaseWiresDrivenComp,
    set_attributes,
)

CaseTwoUpblksSliceComp = set_attributes( CaseTwoUpblksSliceComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Vector4
          port_decl: out Port of Vector8
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
        connections:

        endcomponent
    '''
)

CaseTwoUpblksFreevarsComp = set_attributes( CaseTwoUpblksFreevarsComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
    ''',
    'REF_FREEVAR',
    '''\
        freevars:
          freevar: STATE_IDLE
          freevar: STATE_WORK
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Array[2] of Port
        interface_decls:
        );
        const_decls:
        freevars:
          freevar: STATE_IDLE
          freevar: STATE_WORK
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
        connections:

        endcomponent
    '''
)

CaseBits32TmpWireComp = set_attributes( CaseBits32TmpWireComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Vector32
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseBits32TmpWireAliasComp = set_attributes( CaseBits32TmpWireAliasComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in multi_upblks_1 of Vector32
          tmpvar: u in multi_upblks_2 of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Vector32
          port_decl: out Array[5] of Port
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in multi_upblks_1 of Vector32
          tmpvar: u in multi_upblks_2 of Vector32
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
        connections:

        endcomponent
    '''
)

CaseBits32MultiTmpWireComp = set_attributes( CaseBits32MultiTmpWireComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Vector32
          tmpvar: v in upblk of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Vector32
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Vector32
          tmpvar: v in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseBits32FreeVarToTmpVarComp = set_attributes( CaseBits32FreeVarToTmpVarComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    '''\
        freevars:
          freevar: STATE_IDLE
    ''',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
          freevar: STATE_IDLE
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseBits32ConstBitsToTmpVarComp = set_attributes( CaseBits32ConstBitsToTmpVarComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseBits32ConstIntToTmpVarComp = set_attributes( CaseBits32ConstIntToTmpVarComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseStructTmpWireComp = set_attributes( CaseStructTmpWireComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Struct Bits32Foo
    ''',
    'REF_SRC',
    '''\
        struct Bits32Foo
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Struct Bits32Foo
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Bits32Foo
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseTwoUpblksStructTmpWireComp = set_attributes( CaseTwoUpblksStructTmpWireComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in multi_upblks_1 of Struct Bits32Foo
          tmpvar: u in multi_upblks_2 of Struct Bits32Bar
    ''',
    'REF_SRC',
    '''\
        struct Bits32Foo
        struct Bits32Bar
        component DUT
        (
        port_decls:
          port_decl: in_bar Port of Struct Bits32Bar
          port_decl: in_foo Port of Struct Bits32Foo
          port_decl: out_bar Port of Vector32
          port_decl: out_foo Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in multi_upblks_1 of Bits32Foo
          tmpvar: u in multi_upblks_2 of Bits32Bar
        upblk_srcs:
          upblk_src: multi_upblks_1
          upblk_src: multi_upblks_2
        connections:

        endcomponent
    '''
)

CaseBits32IfcTmpVarOutComp = set_attributes( CaseBits32IfcTmpVarOutComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
          interface_decl: ifc InterfaceView Bits32OutIfc
            interface_ports:
              interface_port: foo Port of Vector32
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseStructIfcTmpVarOutComp = set_attributes( CaseStructIfcTmpVarOutComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Struct Bits32Foo
    ''',
    'REF_SRC',
    '''\
        struct Bits32Foo
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
          interface_decl: ifc InterfaceView Bits32FooInIfc
            interface_ports:
              interface_port: foo Port of Struct Bits32Foo
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Bits32Foo
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseSubCompTmpDrivenComp = set_attributes( CaseSubCompTmpDrivenComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    'freevars:\n',
    'REF_TMPVAR',
    '''\
        tmpvars:
          tmpvar: u in upblk of Vector32
    ''',
    'REF_SRC',
    '''\
        component Bits32OutTmpDrivenComp
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
          tmpvar: u in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent

        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
          component_decl: subcomp Component Bits32OutTmpDrivenComp
            component_ports:
              component_port: out Port of Vector32
            component_ifcs:
        tmpvars:
          tmpvar: u in upblk of Vector32
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
'''
)

CaseSubCompFreeVarDrivenComp = set_attributes( CaseSubCompFreeVarDrivenComp,
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    '''\
        freevars:
          freevar: STATE_IDLE
          freevar: STATE_WORK
    ''',
    'REF_TMPVAR',
    'tmpvars:\n',
    'REF_SRC',
    '''\
        component Bits32OutFreeVarDrivenComp
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
          freevar: STATE_IDLE
          freevar: STATE_WORK
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent

        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
          freevar: STATE_IDLE
        wire_decls:
        component_decls:
          component_decl: subcomp Component Bits32OutFreeVarDrivenComp
            component_ports:
              component_port: out Port of Vector32
            component_ifcs:
        tmpvars:
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseComponentArgsComp = set_attributes( CaseComponentArgsComp,
    'REF_NAME',
    'DUT__foo_0__bar_002a',
    'REF_SRC',
    '''\
        component A__foo_0__bar_002a
        (
        port_decls:
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    '''
)

CaseComponentDefaultArgsComp = set_attributes( CaseComponentDefaultArgsComp,
    'REF_NAME',
    'DUT__foo_0__bar_002a',
    'REF_SRC',
    '''\
        component A__foo_0__bar_002a
        (
        port_decls:
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    '''
)

CaseMixedDefaultArgsComp = set_attributes( CaseMixedDefaultArgsComp,
    'REF_NAME',
    'DUT__foo_0__bar_002a__woo_00000000',
    'REF_SRC',
    '''\
        component A__foo_0__bar_002a__woo_00000000
        (
        port_decls:
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    '''
)

CaseBits32PortOnly = set_attributes( CaseBits32PortOnly,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Port of Vector32
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    'connections:\n',
    'REF_VECTOR',
    [ (rdt.Vector(1), 'Vector1'), (rdt.Vector(32), 'Vector32') ],
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    '''
)

CaseBits32x5PortOnly = set_attributes( CaseBits32x5PortOnly,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Array[5] of Port
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Array[5] of Port
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    '''
)

CaseWiresDrivenComp = set_attributes( CaseWiresDrivenComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    'port_decls:\n',
    'REF_WIRE',
    '''\
        wire_decls:
          wire_decl: bar Wire of Vector4
          wire_decl: foo Wire of Vector32
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
          wire_decl: bar Wire of Vector4
          wire_decl: foo Wire of Vector32
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseBits32Wirex5DrivenComp = set_attributes( CaseBits32Wirex5DrivenComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    'port_decls:\n',
    'REF_WIRE',
    '''\
        wire_decls:
          wire_decl: foo Array[5] of Wire
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
          wire_decl: foo Array[5] of Wire
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    '''
)

CaseBits32ClosureConstruct = set_attributes( CaseBits32ClosureConstruct,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: out Port of Vector32
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    '''\
        const_decls:
          const_decl: fvar_ref Const of Vector32
    ''',
    'REF_SRC',
    # Note that const_decls become emtpy because the constant s.fvar_ref
    # was not used in any upblks!
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
          freevar: foo
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    ''',
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    '''\
        freevars:
          freevar: foo
    '''
)

CaseBits32ArrayClosureConstruct = set_attributes( CaseBits32ArrayClosureConstruct,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: out Port of Vector32
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
          freevar: foo
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    ''',
    'REF_UPBLK',
    '''\
        upblk_srcs:
          upblk_src: upblk
    ''',
    'REF_FREEVAR',
    '''\
        freevars:
          freevar: foo
    '''
)

CaseConnectBitSelToOutComp = set_attributes( CaseConnectBitSelToOutComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Port of Vector32
          port_decl: out Port of Vector1
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    '''\
        connections:
          connection: PartSel CurCompAttr in_ 0 1 -> CurCompAttr out
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Vector32
          port_decl: out Port of Vector1
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: PartSel CurCompAttr in_ 0 1 -> CurCompAttr out

        endcomponent
    '''
)

CaseConnectSliceToOutComp = set_attributes( CaseConnectSliceToOutComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Port of Vector32
          port_decl: out Port of Vector4
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    '''\
        connections:
          connection: PartSel CurCompAttr in_ 4 8 -> CurCompAttr out
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Vector32
          port_decl: out Port of Vector4
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: PartSel CurCompAttr in_ 4 8 -> CurCompAttr out

        endcomponent
    '''
)

CaseConnectPortIndexComp = set_attributes( CaseConnectPortIndexComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Array[5] of Port
          port_decl: out Port of Vector32
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    '''\
        connections:
          connection: PortArrayIdx CurCompAttr in_ 2 -> CurCompAttr out
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Array[5] of Port
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: PortArrayIdx CurCompAttr in_ 2 -> CurCompAttr out

        endcomponent
    '''
)

CaseConnectInToWireComp = set_attributes( CaseConnectInToWireComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Array[5] of Port
          port_decl: out Port of Vector32
    ''',
    'REF_WIRE',
    '''\
        wire_decls:
          wire_decl: wire_ Array[5] of Wire
    ''',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    '''\
        connections:
          connection: WireArrayIdx CurCompAttr wire_ 2 -> CurCompAttr out
          connection: PortArrayIdx CurCompAttr in_ 0 -> WireArrayIdx CurCompAttr wire_ 0
          connection: PortArrayIdx CurCompAttr in_ 1 -> WireArrayIdx CurCompAttr wire_ 1
          connection: PortArrayIdx CurCompAttr in_ 2 -> WireArrayIdx CurCompAttr wire_ 2
          connection: PortArrayIdx CurCompAttr in_ 3 -> WireArrayIdx CurCompAttr wire_ 3
          connection: PortArrayIdx CurCompAttr in_ 4 -> WireArrayIdx CurCompAttr wire_ 4
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: in_ Array[5] of Port
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
          wire_decl: wire_ Array[5] of Wire
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: WireArrayIdx CurCompAttr wire_ 2 -> CurCompAttr out
          connection: PortArrayIdx CurCompAttr in_ 0 -> WireArrayIdx CurCompAttr wire_ 0
          connection: PortArrayIdx CurCompAttr in_ 1 -> WireArrayIdx CurCompAttr wire_ 1
          connection: PortArrayIdx CurCompAttr in_ 2 -> WireArrayIdx CurCompAttr wire_ 2
          connection: PortArrayIdx CurCompAttr in_ 3 -> WireArrayIdx CurCompAttr wire_ 3
          connection: PortArrayIdx CurCompAttr in_ 4 -> WireArrayIdx CurCompAttr wire_ 4

        endcomponent
    '''
)

CaseConnectConstToOutComp = set_attributes( CaseConnectConstToOutComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: out Port of Vector32
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    '''\
        const_decls:
          const_decl: const_ Array[5] of Const
    ''',
    'REF_CONN',
    '''\
        connections:
          connection: Bits32(42) -> CurCompAttr out
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: Bits32(42) -> CurCompAttr out

        endcomponent
    '''
)

CaseStructPortOnly = set_attributes( CaseStructPortOnly,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Port of Struct Bits32Foo
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    'connections:\n',
    'REF_SRC',
    '''\
        struct Bits32Foo
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Struct Bits32Foo
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    ''',
    'REF_STRUCT',
    [(rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}), 'Bits32Foo')]
)

CaseStructWireDrivenComp = set_attributes( CaseStructWireDrivenComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    'port_decls:\n',
    'REF_WIRE',
    '''\
        wire_decls:
          wire_decl: foo Wire of Struct Bits32Foo
    ''',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    'connections:\n',
    'REF_SRC',
    '''\
        struct Bits32Foo
        component DUT
        (
        port_decls:
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
          wire_decl: foo Wire of Struct Bits32Foo
        component_decls:
        tmpvars:
        upblk_srcs:
          upblk_src: upblk
        connections:

        endcomponent
    ''',
    'REF_STRUCT',
    [(rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}), 'Bits32Foo')]
)

CaseStructConstComp = set_attributes( CaseStructConstComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    'port_decls:\n',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    '''\
        const_decls:
          const_decl: struct_const Const of Struct Bits32Foo
    ''',
    'REF_CONN',
    'connections:\n',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    ''',
    'REF_STRUCT',
    [(rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}), 'Bits32Foo')]
)

CaseStructx5PortOnly = set_attributes( CaseStructx5PortOnly,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Array[5] of Port
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    'connections:\n',
    'REF_SRC',
    '''\
        struct Bits32Foo
        component DUT
        (
        port_decls:
          port_decl: in_ Array[5] of Port
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    ''',
    'REF_STRUCT',
    [(rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}), 'Bits32Foo')]
)

CaseNestedStructPortOnly = set_attributes( CaseNestedStructPortOnly,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Port of Struct NestedBits32Foo
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    'connections:\n',
    'REF_SRC',
    '''\
        struct Bits32Foo
        struct NestedBits32Foo
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Struct NestedBits32Foo
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:

        endcomponent
    ''',
    'REF_STRUCT',
    [
      (rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}), 'Bits32Foo'),
      (rdt.Struct('NestedBits32Foo', {'foo':rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)})}), 'NestedBits32Foo'),
    ]
)

CaseNestedPackedArrayStructComp = set_attributes( CaseNestedPackedArrayStructComp,
    'REF_NAME',
    'DUT',
    'REF_PORT',
    '''\
        port_decls:
          port_decl: in_ Port of Struct NestedStructPackedArray
          port_decl: out Port of Struct Bits32x5Foo
    ''',
    'REF_WIRE',
    'wire_decls:\n',
    'REF_CONST',
    'const_decls:\n',
    'REF_CONN',
    '''\
        connections:
          connection: PackedIndex StructAttr CurCompAttr in_ foo 1 -> CurCompAttr out
    ''',
    'REF_SRC',
    '''\
        struct Bits32x5Foo
        struct NestedStructPackedArray
        component DUT
        (
        port_decls:
          port_decl: in_ Port of Struct NestedStructPackedArray
          port_decl: out Port of Struct Bits32x5Foo
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: PackedIndex StructAttr CurCompAttr in_ foo 1 -> CurCompAttr out

        endcomponent
    ''',
    'REF_STRUCT',
    [
      (rdt.Struct('Bits32x5Foo', {'foo':rdt.PackedArray([5], rdt.Vector(32))}), 'Bits32x5Foo'),
      (rdt.Struct('NestedStructPackedArray', {'foo':rdt.Struct('Bits32x5Foo', {'foo':rdt.PackedArray([5], rdt.Vector(32))})}), 'NestedStructPackedArray'),
    ]
)

CaseConnectValRdyIfcComp = set_attributes( CaseConnectValRdyIfcComp,
    'REF_NAME',
    'DUT',
    'REF_IFC',
    '''\
        interface_decls:
          interface_decl: in_ InterfaceView Bits32InValRdyIfc
            interface_ports:
              interface_port: msg Port of Vector32
              interface_port: rdy Port of Vector1
              interface_port: val Port of Vector1
          interface_decl: out InterfaceView Bits32OutValRdyIfc
            interface_ports:
              interface_port: msg Port of Vector32
              interface_port: rdy Port of Vector1
              interface_port: val Port of Vector1
    ''',
    'REF_CONN',
    '''\
        connections:
          connection: IfcAttr CurCompAttr in_ msg -> IfcAttr CurCompAttr out msg
          connection: IfcAttr CurCompAttr out rdy -> IfcAttr CurCompAttr in_ rdy
          connection: IfcAttr CurCompAttr in_ val -> IfcAttr CurCompAttr out val
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
        interface_decls:
          interface_decl: in_ InterfaceView Bits32InValRdyIfc
            interface_ports:
              interface_port: msg Port of Vector32
              interface_port: rdy Port of Vector1
              interface_port: val Port of Vector1
          interface_decl: out InterfaceView Bits32OutValRdyIfc
            interface_ports:
              interface_port: msg Port of Vector32
              interface_port: rdy Port of Vector1
              interface_port: val Port of Vector1
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: IfcAttr CurCompAttr in_ msg -> IfcAttr CurCompAttr out msg
          connection: IfcAttr CurCompAttr out rdy -> IfcAttr CurCompAttr in_ rdy
          connection: IfcAttr CurCompAttr in_ val -> IfcAttr CurCompAttr out val

        endcomponent
    ''',
)

CaseArrayBits32IfcInComp = set_attributes( CaseArrayBits32IfcInComp,
    'REF_NAME',
    'DUT',
    'REF_IFC',
    '''\
        interface_decls:
          interface_decl: in_ Array[2] of InterfaceView Bits32InIfc
            interface_ports:
              interface_port: foo Port of Vector32
    ''',
    'REF_CONN',
    '''\
        connections:
          connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 foo -> CurCompAttr out
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
          interface_decl: in_ Array[2] of InterfaceView Bits32InIfc
            interface_ports:
              interface_port: foo Port of Vector32
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 foo -> CurCompAttr out

        endcomponent
    ''',
)

CaseConnectArrayNestedIfcComp = set_attributes( CaseConnectArrayNestedIfcComp,
    'REF_NAME',
    'DUT',
    'REF_IFC',
    '''\
        interface_decls:
          interface_decl: in_ Array[2] of InterfaceView MemReqIfc
            interface_ports:
              interface_port: ctrl_foo Port of Vector1
              interface_port: memifc InterfaceView ReqIfc
          interface_decl: out Array[2] of InterfaceView MemRespIfc
            interface_ports:
              interface_port: ctrl_foo Port of Vector1
              interface_port: memifc InterfaceView RespIfc
    ''',
    'REF_CONN',
    '''\
        connections:
          connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 ctrl_foo -> IfcAttr IfcArrayIdx CurCompAttr out 0 ctrl_foo
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 0 memifc msg -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 0 memifc msg
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 0 memifc rdy -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 0 memifc rdy
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 0 memifc val -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 0 memifc val
          connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 ctrl_foo -> IfcAttr IfcArrayIdx CurCompAttr out 1 ctrl_foo
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 1 memifc msg -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 1 memifc msg
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 1 memifc rdy -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 1 memifc rdy
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 1 memifc val -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 1 memifc val
    ''',
    'REF_SRC',
    '''\
        component DUT
        (
        port_decls:
        interface_decls:
          interface_decl: in_ Array[2] of InterfaceView MemReqIfc
            interface_ports:
              interface_port: ctrl_foo Port of Vector1
              interface_port: memifc InterfaceView ReqIfc
          interface_decl: out Array[2] of InterfaceView MemRespIfc
            interface_ports:
              interface_port: ctrl_foo Port of Vector1
              interface_port: memifc InterfaceView RespIfc
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 ctrl_foo -> IfcAttr IfcArrayIdx CurCompAttr out 0 ctrl_foo
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 0 memifc msg -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 0 memifc msg
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 0 memifc rdy -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 0 memifc rdy
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 0 memifc val -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 0 memifc val
          connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 ctrl_foo -> IfcAttr IfcArrayIdx CurCompAttr out 1 ctrl_foo
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 1 memifc msg -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 1 memifc msg
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 1 memifc rdy -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 1 memifc rdy
          connection: IfcAttr IfcAttr IfcArrayIdx CurCompAttr in_ 1 memifc val -> IfcAttr IfcAttr IfcArrayIdx CurCompAttr out 1 memifc val

        endcomponent
    ''',
)

CaseBits32ConnectSubCompAttrComp = set_attributes( CaseBits32ConnectSubCompAttrComp,
    'REF_NAME',
    'DUT',
    'REF_CONN',
    '''\
        connections:
          connection: SubCompAttr CurCompAttr b out -> CurCompAttr out
    ''',
    'REF_COMP',
    '''\
        component_decls:
          component_decl: b Component Bits32OutDrivenComp
            component_ports:
              component_port: out Port of Vector32
            component_ifcs:
    ''',
    'REF_SRC',
    '''\
        component Bits32OutDrivenComp
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: Bits32(42) -> CurCompAttr out

        endcomponent

        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
          component_decl: b Component Bits32OutDrivenComp
            component_ports:
              component_port: out Port of Vector32
            component_ifcs:
        tmpvars:
        upblk_srcs:
        connections:
          connection: SubCompAttr CurCompAttr b out -> CurCompAttr out

        endcomponent
    ''',
)

CaseConnectSubCompIfcHierarchyComp = set_attributes( CaseConnectSubCompIfcHierarchyComp,
    'REF_NAME',
    'DUT',
    'REF_CONN',
    '''\
        connections:
          connection: SubCompAttr CurCompAttr subcomp out -> CurCompAttr out
          connection: IfcAttr SubCompAttr CurCompAttr subcomp ifc msg -> IfcAttr CurCompAttr ifc msg
          connection: IfcAttr CurCompAttr ifc rdy -> IfcAttr SubCompAttr CurCompAttr subcomp ifc rdy
          connection: IfcAttr SubCompAttr CurCompAttr subcomp ifc val -> IfcAttr CurCompAttr ifc val
    ''',
    'REF_COMP',
    '''\
        component_decls:
          component_decl: subcomp Component Bits32OutDrivenSubComp
            component_ports:
              component_port: out Port of Vector32
            component_ifcs:
              component_ifc: ifc InterfaceView Bits32OutValRdyIfc
                component_ifc_ports:
                  component_ifc_port: msg Port of Vector32
                  component_ifc_port: rdy Port of Vector1
                  component_ifc_port: val Port of Vector1
    ''',
    'REF_SRC',
    '''\
        component Bits32OutDrivenSubComp
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
          interface_decl: ifc InterfaceView Bits32OutValRdyIfc
            interface_ports:
              interface_port: msg Port of Vector32
              interface_port: rdy Port of Vector1
              interface_port: val Port of Vector1
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
        tmpvars:
        upblk_srcs:
        connections:
          connection: Bits32(42) -> CurCompAttr out
          connection: Bits32(42) -> IfcAttr CurCompAttr ifc msg
          connection: Bits1(1) -> IfcAttr CurCompAttr ifc val

        endcomponent

        component DUT
        (
        port_decls:
          port_decl: out Port of Vector32
        interface_decls:
          interface_decl: ifc InterfaceView Bits32OutValRdyIfc
            interface_ports:
              interface_port: msg Port of Vector32
              interface_port: rdy Port of Vector1
              interface_port: val Port of Vector1
        );
        const_decls:
        freevars:
        wire_decls:
        component_decls:
          component_decl: subcomp Component Bits32OutDrivenSubComp
            component_ports:
              component_port: out Port of Vector32
            component_ifcs:
              component_ifc: ifc InterfaceView Bits32OutValRdyIfc
                component_ifc_ports:
                  component_ifc_port: msg Port of Vector32
                  component_ifc_port: rdy Port of Vector1
                  component_ifc_port: val Port of Vector1
        tmpvars:
        upblk_srcs:
        connections:
          connection: SubCompAttr CurCompAttr subcomp out -> CurCompAttr out
          connection: IfcAttr SubCompAttr CurCompAttr subcomp ifc msg -> IfcAttr CurCompAttr ifc msg
          connection: IfcAttr CurCompAttr ifc rdy -> IfcAttr SubCompAttr CurCompAttr subcomp ifc rdy
          connection: IfcAttr SubCompAttr CurCompAttr subcomp ifc val -> IfcAttr CurCompAttr ifc val

        endcomponent
    ''',
)



#=========================================================================
# BehavioralRTLIRTypeCheckL4Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L4 behavioral RTLIR type check pass."""

from collections import OrderedDict

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL3Pass import BehavioralRTLIRTypeCheckVisitorL3


class BehavioralRTLIRTypeCheckL4Pass( BasePass ):
  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    if not hasattr( m, '_pass_behavioral_rtlir_type_check' ):
      m._pass_behavioral_rtlir_type_check = PassMetadata()
    m._pass_behavioral_rtlir_type_check.rtlir_freevars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_tmpvars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_accessed = set()
    visitor = BehavioralRTLIRTypeCheckVisitorL4(
      m,
      m._pass_behavioral_rtlir_type_check.rtlir_freevars,
      m._pass_behavioral_rtlir_type_check.rtlir_accessed,
      m._pass_behavioral_rtlir_type_check.rtlir_tmpvars
    )

    for blk in m.get_update_block_order():
      visitor.enter( blk, m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )

class BehavioralRTLIRTypeCheckVisitorL4( BehavioralRTLIRTypeCheckVisitorL3 ):
  def __init__( s, component, freevars, accessed, tmpvars ):
    super(). \
        __init__( component, freevars, accessed, tmpvars )
    s.type_expect[ 'Attribute' ] = {
      'value':( (rt.Component, rt.Signal, rt.InterfaceView),
        'the base of an attribute must be one of: module, struct, interface!' )
    }

  def visit_Attribute( s, node ):
    if isinstance( node.value.Type, rt.InterfaceView ):
      if not node.value.Type.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{dtype.get_name()} does not have field {node.attr}!' )
      node.Type = node.value.Type.get_property( node.attr )
    else:
      super().visit_Attribute( node )

  def visit_Index( s, node ):
    if isinstance( node.value.Type, rt.Array ) and \
       isinstance( node.value.Type.get_sub_type(), rt.InterfaceView ):
      node.Type = node.value.Type.get_sub_type()

    else:
      super().visit_Index( node )



#=========================================================================
# BehavioralRTLIRVisualizationPass.py
#=========================================================================
"""Provide visualization for behavioral RTLIR AST.

Visualize Behavioral RTLIR using Graphviz packeage. The output graph is in PDF
format.  This file is automatically generated by BehavioralRTLIRImplGen.py.
"""

import os

from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.rtlir.rtype.RTLIRType import BaseRTLIRType

from .BehavioralRTLIR import BehavioralRTLIRNodeVisitor


class BehavioralRTLIRVisualizationPass( BasePass ):
  def __call__( s, model ):
    visitor = BehavioralRTLIRVisualizationVisitor()

    for blk in model.get_update_blocks():
      visitor.init( blk.__name__ )
      visitor.visit( model._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )
      visitor.dump()

class BehavioralRTLIRVisualizationVisitor( BehavioralRTLIRNodeVisitor ):
  def __init__( s ):
    s.output = 'unamed'
    s.output_dir = 'rast-viz'
    s.table_header = '<<TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0"> '
    s.table_trail = ' </TABLE>>'

  def init( s, name ):
    from graphviz import Digraph
    s.g = Digraph(
      comment = 'BehavioralRTLIR Visualization of ' + name,
      node_attr = { 'shape' : 'plaintext' }
    )
    s.blk_name = name
    s.cur = 0

  def get_str( s, obj ):
    return str(obj).replace('<', '&lt;').replace('>', '&gt;')

  def gen_table_opt( s, node ):
    ret = ''
    if isinstance( node.Type, BaseRTLIRType ):
      ret = ' <TR><TD COLSPAN="2">Type: ' + node.Type.__class__.__name__ + '</TD></TR>'
      for name, obj in vars(node.Type).items():
        obj_str = s.get_str( obj )
        if not isinstance( obj, dict ):
          ret += ' <TR><TD>' + name + '</TD><TD>' + obj_str + '</TD></TR>'
        else:
          ret += ' <TR><TD>' + name + '</TD><TD>{' + obj_str + '}</TD></TR>'
    return ret

  def dump( s ):
    if not os.path.exists( s.output_dir ):
      os.makedirs( s.output_dir )
    s.g.render( s.output_dir + os.sep + s.blk_name )

  def visit_CombUpblk( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">CombUpblk</TD></TR> <TR><TD>name</TD><TD>{name}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(name=s.get_str(node.name))
    s.g.node( str( s.cur ), label = label )
    for i, f in enumerate(node.body):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'body[{idx}]'.format(idx = i) )
      s.visit( f )

  def visit_SeqUpblk( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">SeqUpblk</TD></TR> <TR><TD>name</TD><TD>{name}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(name=s.get_str(node.name))
    s.g.node( str( s.cur ), label = label )
    for i, f in enumerate(node.body):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'body[{idx}]'.format(idx = i) )
      s.visit( f )

  def visit_Assign( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Assign</TD></TR> <TR><TD>blocking</TD><TD>{blocking}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(blocking=s.get_str(node.blocking))
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'target' )
    s.visit( node.target )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )

  def visit_If( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">If</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'cond' )
    s.visit( node.cond )
    for i, f in enumerate(node.body):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'body[{idx}]'.format(idx = i) )
      s.visit( f )
    for i, f in enumerate(node.orelse):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'orelse[{idx}]'.format(idx = i) )
      s.visit( f )

  def visit_For( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">For</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'var' )
    s.visit( node.var )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'start' )
    s.visit( node.start )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'end' )
    s.visit( node.end )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'step' )
    s.visit( node.step )
    for i, f in enumerate(node.body):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'body[{idx}]'.format(idx = i) )
      s.visit( f )

  def visit_Number( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Number</TD></TR> <TR><TD>value</TD><TD>{value}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(value=s.get_str(node.value))
    s.g.node( str( s.cur ), label = label )

  def visit_Concat( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Concat</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    for i, f in enumerate(node.values):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'values[{idx}]'.format(idx = i) )
      s.visit( f )

  def visit_ZeroExt( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">ZeroExt</TD></TR> <TR><TD>nbits</TD><TD>{nbits}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(nbits=s.get_str(node.nbits))
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )

  def visit_SignExt( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">SignExt</TD></TR> <TR><TD>nbits</TD><TD>{nbits}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(nbits=s.get_str(node.nbits))
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )

  def visit_Reduce( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Reduce</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'op' )
    s.visit( node.op )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )

  def visit_SizeCast( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">SizeCast</TD></TR> <TR><TD>nbits</TD><TD>{nbits}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(nbits=s.get_str(node.nbits))
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )

  def visit_StructInst( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">StructInst</TD></TR> <TR><TD>struct</TD><TD>{struct}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(struct=s.get_str(node.struct))
    s.g.node( str( s.cur ), label = label )
    for i, f in enumerate(node.values):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'values[{idx}]'.format(idx = i) )
      s.visit( f )

  def visit_IfExp( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">IfExp</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'cond' )
    s.visit( node.cond )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'body' )
    s.visit( node.body )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'orelse' )
    s.visit( node.orelse )

  def visit_UnaryOp( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">UnaryOp</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'op' )
    s.visit( node.op )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'operand' )
    s.visit( node.operand )

  def visit_BoolOp( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">BoolOp</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'op' )
    s.visit( node.op )
    for i, f in enumerate(node.values):
      s.g.edge( str(local_cur), str(s.cur+1), label = 'values[{idx}]'.format(idx = i) )
      s.visit( f )

  def visit_BinOp( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">BinOp</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'left' )
    s.visit( node.left )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'op' )
    s.visit( node.op )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'right' )
    s.visit( node.right )

  def visit_Compare( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Compare</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'left' )
    s.visit( node.left )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'op' )
    s.visit( node.op )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'right' )
    s.visit( node.right )

  def visit_Attribute( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Attribute</TD></TR> <TR><TD>attr</TD><TD>{attr}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(attr=s.get_str(node.attr))
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )

  def visit_Index( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Index</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'idx' )
    s.visit( node.idx )

  def visit_Slice( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Slice</TD></TR> <TR><TD>size</TD><TD>{size}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(size=s.get_str(node.size))
    s.g.node( str( s.cur ), label = label )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'value' )
    s.visit( node.value )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'lower' )
    s.visit( node.lower )
    s.g.edge( str(local_cur), str(s.cur+1), label = 'upper' )
    s.visit( node.upper )
    if node.base is not None:
      s.g.edge( str(local_cur), str(s.cur+1), label = 'base' )
      s.visit( node.base )

  def visit_Base( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Base</TD></TR> <TR><TD>base</TD><TD>{base}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(base=s.get_str(node.base))
    s.g.node( str( s.cur ), label = label )

  def visit_LoopVar( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">LoopVar</TD></TR> <TR><TD>name</TD><TD>{name}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(name=s.get_str(node.name))
    s.g.node( str( s.cur ), label = label )

  def visit_FreeVar( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">FreeVar</TD></TR> <TR><TD>name</TD><TD>{name}</TD></TR> <TR><TD>obj</TD><TD>{obj}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(name=s.get_str(node.name), obj=s.get_str(node.obj))
    s.g.node( str( s.cur ), label = label )

  def visit_TmpVar( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">TmpVar</TD></TR> <TR><TD>name</TD><TD>{name}</TD></TR> <TR><TD>upblk_name</TD><TD>{upblk_name}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(name=s.get_str(node.name), upblk_name=s.get_str(node.upblk_name))
    s.g.node( str( s.cur ), label = label )

  def visit_LoopVarDecl( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">LoopVarDecl</TD></TR> <TR><TD>name</TD><TD>{name}</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail).format(name=s.get_str(node.name))
    s.g.node( str( s.cur ), label = label )

  def visit_Invert( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Invert</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Not( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Not</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_UAdd( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">UAdd</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_USub( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">USub</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_And( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">And</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Or( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Or</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Add( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Add</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Sub( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Sub</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Mult( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Mult</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Div( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Div</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Mod( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Mod</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Pow( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Pow</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_ShiftLeft( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">ShiftLeft</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_ShiftRightLogic( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">ShiftRightLogic</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_BitAnd( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">BitAnd</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_BitOr( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">BitOr</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_BitXor( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">BitXor</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Eq( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Eq</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_NotEq( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">NotEq</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Lt( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Lt</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_LtE( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">LtE</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_Gt( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">Gt</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )

  def visit_GtE( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '<TR><TD COLSPAN="2">GtE</TD></TR>'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail)
    s.g.node( str( s.cur ), label = label )



#=========================================================================
# BehavioralTranslatorL2.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 22, 2019
"""Provide L2 behavioral translator."""

from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL2Pass import (
    BehavioralRTLIRGenL2Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL2Pass import (
    BehavioralRTLIRTypeCheckL2Pass,
)

from .BehavioralTranslatorL1 import BehavioralTranslatorL1


class BehavioralTranslatorL2( BehavioralTranslatorL1 ):
  def __init__( s, top ):
    super().__init__( top )

  def clear( s, tr_top ):
    super().clear( tr_top )

  #-----------------------------------------------------------------------
  # gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def gen_behavioral_trans_metadata( s, tr_top ):
    s.behavioral.tmpvars = {}
    s.behavioral.decl_tmpvars = {}
    super().gen_behavioral_trans_metadata( tr_top )

  #-----------------------------------------------------------------------
  # _gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def _gen_behavioral_trans_metadata( s, m ):
    m.apply( BehavioralRTLIRGenL2Pass() )
    m.apply( BehavioralRTLIRTypeCheckL2Pass() )
    s.behavioral.rtlir[m] = m._pass_behavioral_rtlir_gen.rtlir_upblks
    s.behavioral.freevars[m] = \
        m._pass_behavioral_rtlir_type_check.rtlir_freevars
    s.behavioral.tmpvars[m] = \
        m._pass_behavioral_rtlir_type_check.rtlir_tmpvars

  #-----------------------------------------------------------------------
  # translate_behavioral
  #-----------------------------------------------------------------------

  # Override
  def translate_behavioral( s, m ):
    """Translate behavioral part of `m`.

    Support for translating temporary variables is added at level 2.
    """
    super().translate_behavioral(m)

    # Generate temporary variable declarations
    tmpvars = []
    for (id_, upblk_id), rtype in s.behavioral.tmpvars[m].items():
      assert isinstance(rtype, rt.Wire), \
        f"temporary variable {id_} in upblk {upblk_id} is not a signal!"
      dtype = rtype.get_dtype()
      tmpvars.append( s.rtlir_tr_behavioral_tmpvar(
        id_,
        upblk_id,
        s.rtlir_data_type_translation( m, dtype )
      ) )
    s.behavioral.decl_tmpvars[m] = s.rtlir_tr_behavioral_tmpvars( tmpvars )

  #-----------------------------------------------------------------------
  # Methods to be implemented by the backend translator
  #-----------------------------------------------------------------------

  def rtlir_tr_behavioral_tmpvars( s, tmpvars ):
    raise NotImplementedError()

  def rtlir_tr_behavioral_tmpvar( s, name, upblk_name, dtype ):
    raise NotImplementedError()



#=========================================================================
# StructuralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 1 structural translators."""

import pytest

from pymtl3 import Bits4, Bits16
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.util.test_utility import expected_failure

from ...testcases import (
    CaseBits32ArrayClosureConstruct,
    CaseBits32ClosureConstruct,
    CaseBits32PortOnly,
    CaseBits32Wirex5DrivenComp,
    CaseBits32x5PortOnly,
    CaseComponentArgsComp,
    CaseComponentDefaultArgsComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectPortIndexComp,
    CaseConnectSliceToOutComp,
    CaseDoubleStarArgComp,
    CaseMixedDefaultArgsComp,
    CaseStarArgComp,
    CaseWiresDrivenComp,
)
from ..StructuralTranslatorL1 import StructuralTranslatorL1
from .TestStructuralTranslator import mk_TestStructuralTranslator


def run_test( case, m ):
  m.elaborate()
  tr = mk_TestStructuralTranslator(StructuralTranslatorL1)(m)
  tr.clear( m )
  tr.translate_structural(m)
  try:
    name = tr.structural.component_unique_name[m]
    assert name == case.REF_NAME
    decl_ports = tr.structural.decl_ports[m]
    assert decl_ports == case.REF_PORT
    decl_wires = tr.structural.decl_wires[m]
    assert decl_wires == case.REF_WIRE
    decl_consts = tr.structural.decl_consts[m]
    assert decl_consts == case.REF_CONST
    connections = tr.structural.connections[m]
    assert connections == case.REF_CONN
    vector_types = tr.structural.decl_type_vector
    assert sorted(vector_types, key=lambda x: str(x[0])) == case.REF_VECTOR
  except AttributeError:
    pass

def test_component_args():
  case = CaseComponentArgsComp
  run_test( case, case.DUT( Bits4(0), Bits16(42) ))

def test_component_default_args():
  case = CaseComponentDefaultArgsComp
  run_test( case, case.DUT( Bits4(0) ) )

def test_component_kw_args():
  case = CaseComponentArgsComp
  run_test( case, case.DUT( foo = Bits4(0), bar = Bits16(42) ) )

def test_component_star_args():
  args = [ Bits4(0), Bits16(42) ]
  case = CaseStarArgComp
  with expected_failure( RTLIRConversionError, "varargs are not allowed" ):
    run_test( case, case.DUT( *args ) )

def test_component_star_args_ungroup():
  args = [ Bits4(0), Bits16(42) ]
  case = CaseComponentArgsComp
  run_test( case, case.DUT( *args ) )

def test_component_double_star_args():
  kwargs = { 'foo':Bits4(0), 'bar':Bits16(42) }
  case = CaseDoubleStarArgComp
  with expected_failure( RTLIRConversionError, "keyword args are not allowed" ):
    run_test( case, case.DUT( **kwargs ) )

def test_component_double_star_args_ungroup():
  kwargs = { 'foo':Bits4(0), 'bar':Bits16(42) }
  case = CaseComponentArgsComp
  run_test( case, case.DUT( **kwargs ) )

def test_component_mixed_kw_args():
  case = CaseMixedDefaultArgsComp
  run_test( case, case.DUT( Bits4(0), bar = Bits16(42) ) )

@pytest.mark.parametrize(
  'case', [
    CaseBits32PortOnly,
    CaseBits32x5PortOnly,
    CaseWiresDrivenComp,
    CaseBits32Wirex5DrivenComp,
    CaseBits32ClosureConstruct,
    CaseBits32ArrayClosureConstruct,
    CaseConnectBitSelToOutComp,
    CaseConnectSliceToOutComp,
    CaseConnectPortIndexComp,
    CaseConnectInToWireComp,
    CaseConnectConstToOutComp,
  ]
)
def test_generic_structural_L1( case ):
  run_test( case, case.DUT() )



#=========================================================================
# YosysTranslator_L3_cases_test.py
#=========================================================================
"""Test the yosys-SystemVerilog translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.rtlir.util.test_utility import get_parameter

from ..behavioral.test.YosysBehavioralTranslatorL4_test import test_yosys_behavioral_L4
from ..structural.test.YosysStructuralTranslatorL3_test import test_yosys_structural_L3
from ..YosysTranslator import YosysTranslator


def run_test( case, m ):
  m.elaborate()
  tr = YosysTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, case.REF_SRC )

@pytest.mark.parametrize(
  'case', get_parameter('case', test_yosys_behavioral_L4) + \
          get_parameter('case', test_yosys_structural_L3)
)
def test_yosys_L3( case ):
  run_test( case, case.DUT() )



#=========================================================================
# YosysBehavioralTranslatorL5.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog L5 behavioral translator."""

from pymtl3.passes.backends.sverilog.translation.behavioral.SVBehavioralTranslatorL5 import (
    BehavioralRTLIRToSVVisitorL5,
    SVBehavioralTranslatorL5,
)
from pymtl3.passes.rtlir import BehavioralRTLIR as bir
from pymtl3.passes.rtlir import RTLIRType as rt

from .YosysBehavioralTranslatorL4 import (
    YosysBehavioralRTLIRToSVVisitorL4,
    YosysBehavioralTranslatorL4,
)


class YosysBehavioralTranslatorL5(
    YosysBehavioralTranslatorL4, SVBehavioralTranslatorL5 ):

  def _get_rtlir2sv_visitor( s ):
    return YosysBehavioralRTLIRToSVVisitorL5

class YosysBehavioralRTLIRToSVVisitorL5(
    YosysBehavioralRTLIRToSVVisitorL4, BehavioralRTLIRToSVVisitorL5 ):

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for subcomponent attributes in L5.
    """
    # Generate subcomponent attribute
    if isinstance( node.value.Type, rt.Component ) and\
       not isinstance( node.value, bir.Base ):

      value = s.visit( node.value )
      s.signal_expr_prologue( node )
      attr = node.attr
      s.check_res( node, attr )
      node.sexpr['s_attr'] += "__{}"
      node.sexpr['attr'].append( attr )
      return s.signal_expr_epilogue(node, f"{value}.{attr}")

    return super().visit_Attribute( node )



#=========================================================================
# utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 27, 2019
"""Provide helper methods that might be useful to sverilog passes."""

import os
import shutil
import textwrap
from hashlib import blake2b

from pymtl3.passes.rtlir.util.utility import get_component_full_name


def make_indent( src, nindent ):
  """Add nindent indention to every line in src."""
  indent = '  '
  for idx, s in enumerate( src ):
    src[ idx ] = nindent * indent + s

def get_component_unique_name( c_rtype ):
  full_name = get_component_full_name( c_rtype )

  if len( full_name ) < 64:
    return full_name

  comp_name = c_rtype.get_name()
  param_hash = blake2b(digest_size = 8)
  param_hash.update(full_name[len(comp_name):].encode('ascii'))
  param_name = param_hash.hexdigest()
  return comp_name + "__" + param_name

  def get_string( obj ):
    """Return the string that identifies `obj`"""
    if isinstance(obj, type): return obj.__name__
    return str( obj )

def wrap( s ):
  col = shutil.get_terminal_size().columns
  return "\n".join(sum((textwrap.wrap(line, col) for line in s.split("\n")), []))

def expand( v ):
  return os.path.expanduser(os.path.expandvars(v))

def get_dir( cur_file ):
  return os.path.dirname(os.path.abspath(cur_file))+os.path.sep

def get_file_hash( file_path ):
  with open( file_path, 'r' ) as fd:
    hash_inst = blake2b()
    string = ''.join( fd.readlines() ).encode( 'ascii' )
    hash_inst.update(string)
    return hash_inst.hexdigest()

sverilog_keyword = [
  # Verilog-1995 reserved keywords
  "always", "and", "assign", "begin", "buf", "bufif0", "bufif1", "case",
  "casex", "casez", "cmos", "deassign", "default", "defparam", "disable",
  "edge", "else", "end", "endcase", "endmodule", "endfunction", "endprimitive",
  "endspecify", "endtable", "endtask", "event", "for", "force", "forever",
  "fork", "function", "highz0", "highz1", "if", "ifnone", "initial",
  "inout", "input", "output", "integer", "join", "large", "macromodule",
  "medium", "module", "nand", "negedge", "nmos", "nor", "not", "notif0",
  "notif1", "or", "output", "parameter", "pmos", "posedge", "primitive",
  "pull0", "pull1", "pullup", "pulldown", "rcmos", "real", "realtime",
  "reg", "release", "repeat", "rnmos", "rpmos", "rtran", "rtranif0",
  "rtranif1", "scalared", "small", "specify", "specparam", "strong0",
  "strong1", "supply0", "supply1", "table", "task", "time", "tran",
  "tranif0", "tranif1", "tri", "tri0", "tri1", "triand", "trior",
  "trireg", "vectored", "wait", "wand", "weak0", "weak1", "while",
  "wire", "wor", "xnor", "xor",
  # Verilog-2001 reserved keywords
  "automatic", "cell", "config", "design", "endconfig", "endgenerate",
  "generate", "genvar", "incdir", "include", "instance", "liblist",
  "library", "localparam", "noshowcancelled", "pulsestyle_onevent",
  "pulsestyle_ondetect", "showcancelled", "signed", "unsigned", "use",
  # Verilog-2005 reserved keywords
  "uwire",
  # SystemVerilog-2005 reserved keywords
  "alias", "always_comb", "always_ff", "always_latch", "assert", "assume",
  "before", "bind", "bins", "binsof", "bit", "break", "byte", "chandle",
  "class", "clocking", "const", "constraint", "context", "continue",
  "cover", "covergroup", "coverpoint", "cross", "dist", "do", "endclass",
  "endclocking", "endgroup", "endinterface", "endpackage", "endprimitive",
  "endprogram", "endproperty", "endsequence", "enum", "expect", "export",
  "extends", "extern", "final", "first_match", "foreach", "forkjoin",
  "iff", "ignore_bins", "illegal_bins", "import", "inside", "int", "interface",
  "intersect", "join_any", "join_none", "local", "logic", "longint", "matches",
  "modport", "new", "null", "package", "packed", "priority", "program",
  "property", "protected", "pure", "rand", "randc", "randcase", "randsequence",
  "ref", "return", "sequence", "shortint", "shortreal", "solve", "static",
  "string", "struct", "super", "tagged", "this", "throughout", "timeprecision",
  "timeunit", "type", "typedef", "union", "unique", "var", "virtual", "void",
  "wait_order", "wildcard", "with", "within"
]

sverilog_reserved = set( sverilog_keyword )



#=========================================================================
# BehavioralRTLIRL3Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the level 3 behavioral RTLIR passes.

The L3 generation, L3 type check, and visualization passes are invoked. The
generation pass results are verified against a reference AST.
"""

from pymtl3.dsl.errors import VarNotDeclaredError
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIR import *
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL3Pass import (
    BehavioralRTLIRGenL3Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL3Pass import (
    BehavioralRTLIRTypeCheckL3Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRVisualizationPass import (
    BehavioralRTLIRVisualizationPass,
)
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError, PyMTLTypeError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    Bits32Foo,
    CaseBits32FooInBits32OutComp,
    CaseBits32FooInstantiationComp,
    CaseBits32FooKwargComp,
    CaseBitsAttributeComp,
    CaseConstStructInstComp,
    CaseStructMissingAttributeComp,
)


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  if isinstance(m, type):
    m = m.DUT()
  m.elaborate()
  m.apply( BehavioralRTLIRGenL3Pass() )
  m.apply( BehavioralRTLIRTypeCheckL3Pass() )
  m.apply( BehavioralRTLIRVisualizationPass() )

  try:
    ref = m._rtlir_test_ref
    for blk in m.get_update_blocks():
      upblk = m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ]
      assert upblk == ref[ blk.__name__ ]
  except AttributeError:
    pass

#-------------------------------------------------------------------------
# Correct test cases
#-------------------------------------------------------------------------

def test_L3_struct_attr( do_test ):
  a = CaseBits32FooInBits32OutComp.DUT()
  a._rtlir_test_ref = { 'upblk' : CombUpblk( 'upblk', [ Assign(
      Attribute( Base( a ), 'out' ), Attribute(
        Attribute( Base( a ), 'in_' ), 'foo' ), True ) ] ) }
  do_test( a )

def test_L3_struct_inst_kwargs( do_test ):
  a = CaseBits32FooKwargComp.DUT()
  a._rtlir_test_ref = { 'upblk' : CombUpblk( 'upblk', [ Assign(
      Attribute( Base( a ), 'out' ), StructInst(
        Bits32Foo, [ SizeCast( 32, Number( 42 ) ) ] ), True ) ] ) }
  with expected_failure( PyMTLSyntaxError, 'keyword argument is not supported' ):
    do_test( a )

def test_L3_struct_inst( do_test ):
  a = CaseBits32FooInstantiationComp.DUT()
  a._rtlir_test_ref = { 'upblk' : CombUpblk( 'upblk', [ Assign(
      Attribute( Base( a ), 'out' ), StructInst(
        Bits32Foo, [ SizeCast( 32, Number( 42 ) ) ] ), True ) ] ) }
  do_test( a )

def test_L3_const_struct( do_test ):
  a = CaseConstStructInstComp.DUT()
  a._rtlir_test_ref = { 'upblk' : CombUpblk( 'upblk', [ Assign(
      Attribute( Base( a ), 'out' ), Attribute(
        Attribute( Base( a ), 'in_' ), 'foo' ), True ) ] ) }
  do_test( a )

#-------------------------------------------------------------------------
# PyMTL type errors
#-------------------------------------------------------------------------

def test_L3_vector_attr( do_test ):
  with expected_failure( VarNotDeclaredError, 's.in_ does not have field "foo"' ):
    do_test( CaseBitsAttributeComp )

def test_L3_struct_no_field( do_test ):
  with expected_failure( VarNotDeclaredError, 's.in_ does not have field "bar"' ):
    do_test( CaseStructMissingAttributeComp )



#=========================================================================
# errors.py
#=========================================================================
# Author : Shunning Jiang, Peitian Pan
# Date   : Jan 4, 2019
"""Behavioral code errors."""

import inspect
import os
import sys
import traceback


class RTLIRConversionError( Exception ):
  """Error while converting a PyMTL structure into RTLIR."""
  def __init__( self, obj, msg ):
    obj = str(obj)
    _, _, tb = sys.exc_info()
    traceback.print_tb(tb)
    tb_info = traceback.extract_tb(tb)
    fname, line, func, text = tb_info[-1]
    return super().__init__(
      f"\nIn file {fname}, Line {line}, Method {func}:"
      f"\nError trying to convert {obj} into RTLIR:\n- {msg}"
      f"\n  {text}" )

class PyMTLSyntaxError( Exception ):
  """Behavioral RTLIR syntax error."""
  def __init__( self, blk, ast, msg ):
    fname = os.path.abspath( inspect.getsourcefile( blk ) )
    line = inspect.getsourcelines( blk )[1]
    col = 0
    code = ""
    try:
      line += ast.lineno - 1
      col = ast.col_offset
      code_line = inspect.getsourcelines( blk )[0][ ast.lineno-1 ]
      code = '\n  ' + code_line.strip() + \
        '\n  '+ ' ' * (col-len(code_line)+len(code_line.lstrip())) + '^'
    except AttributeError:
      # The given AST node is neither expr nor stmt
      pass
    return super().__init__(
      f"\nIn file {fname}, Line {line}, Col {col}:{code}\n- {msg}" )

class PyMTLTypeError( Exception ):
  """Behavioral RTLIR type error."""
  def __init__( self, blk, ast, msg ):
    fname = os.path.abspath( inspect.getsourcefile( blk ) )
    line = inspect.getsourcelines( blk )[1]
    col = 0
    code = ""
    try:
      line += ast.lineno - 1
      col = ast.col_offset
      code_line = inspect.getsourcelines( blk )[0][ ast.lineno-1 ]
      code = '\n  ' + code_line.strip() + \
        '\n  '+ ' ' * (col-len(code_line)+len(code_line.lstrip())) + '^'
    except AttributeError:
      # The given AST node is neither expr nor stmt
      pass
    return super().__init__(
      f"\nIn file {fname}, Line {line}, Col {col}:{code}\n- {msg}" )



#=========================================================================
# StructuralRTLIRGenL3Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 19, 2019
"""Test the generation of level 1 structural RTLIR."""

from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL3Pass import (
    StructuralRTLIRGenL3Pass,
)
from pymtl3.passes.rtlir.structural.StructuralRTLIRSignalExpr import *
from pymtl3.passes.testcases import (
    CaseArrayBits32IfcInComp,
    CaseBits32IfcInComp,
    CaseConnectValRdyIfcComp,
)

from .StructuralRTLIRGenL1Pass_test import gen_connections


def test_L3_ifc_view_attr():
  a = CaseBits32IfcInComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL3Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  assert ns.connections == \
    [(InterfaceAttr(CurCompAttr(comp, 'in_'), 'foo'), CurCompAttr(comp, 'out'))]

def test_L3_ifc_view_index():
  a = CaseArrayBits32IfcInComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL3Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  assert ns.connections == \
    [(InterfaceAttr(InterfaceViewIndex(CurCompAttr(comp, 'in_'), 1), 'foo'),
     CurCompAttr(comp, 'out'))]

def test_L3_ifc_view_connection():
  a = CaseConnectValRdyIfcComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL3Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  ref = \
    [
      (InterfaceAttr(CurCompAttr(comp, 'in_'), 'msg'),
      InterfaceAttr(CurCompAttr(comp, 'out'), 'msg')),
      (InterfaceAttr(CurCompAttr(comp, 'in_'), 'val'),
      InterfaceAttr(CurCompAttr(comp, 'out'), 'val')),
      (InterfaceAttr(CurCompAttr(comp, 'out'), 'rdy'),
      InterfaceAttr(CurCompAttr(comp, 'in_'), 'rdy')),
    ]
  # The order of ports is non-deterministic?
  assert ns.connections[0] in ref
  assert ns.connections[1] in ref
  assert ns.connections[2] in ref



#=========================================================================
# YosysBehavioralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Test the YosysVerilog translator implementation."""

import pytest

from pymtl3.passes.backends.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import expected_failure

from ....testcases import (
    CaseBits32BitSelUpblkComp,
    CaseBits32x2ConcatComp,
    CaseBits32x2ConcatConstComp,
    CaseBits32x2ConcatFreeVarComp,
    CaseBits32x2ConcatMixedComp,
    CaseBits32x2ConcatUnpackedSignalComp,
    CaseBits64PartSelUpblkComp,
    CaseBits64SextInComp,
    CaseBits64ZextInComp,
    CasePassThroughComp,
    CaseSequentialPassThroughComp,
    CaseSVerilogReservedComp,
)
from ..YosysBehavioralTranslatorL1 import YosysBehavioralRTLIRToSVVisitorL1


def run_test( case, m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = YosysBehavioralRTLIRToSVVisitorL1(lambda x: x in sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  assert len(m_all_upblks) == 1

  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src + '\n' == case.REF_UPBLK

@pytest.mark.parametrize(
    'case', [
      CasePassThroughComp,
      CaseSequentialPassThroughComp,
      CaseBits32x2ConcatComp,
      CaseBits32x2ConcatConstComp,
      CaseBits32x2ConcatMixedComp,
      CaseBits64SextInComp,
      CaseBits64ZextInComp,
      CaseBits32x2ConcatFreeVarComp,
      CaseBits32x2ConcatUnpackedSignalComp,
      CaseBits32BitSelUpblkComp,
      CaseBits64PartSelUpblkComp,
    ]
)
def test_yosys_behavioral_L1( case ):
  run_test( case, case.DUT() )

def test_yosys_reserved_keyword():
  with expected_failure( SVerilogTranslationError, "reserved keyword" ):
    run_test( CaseSVerilogReservedComp, CaseSVerilogReservedComp.DUT() )



#=========================================================================
# SVBehavioralTranslatorL5.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 5 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL5 import (
    BehavioralTranslatorL5,
)
from pymtl3.passes.rtlir import BehavioralRTLIR as bir
from pymtl3.passes.rtlir import RTLIRType as rt

from ...errors import SVerilogTranslationError
from .SVBehavioralTranslatorL4 import (
    BehavioralRTLIRToSVVisitorL4,
    SVBehavioralTranslatorL4,
)


class SVBehavioralTranslatorL5(
    SVBehavioralTranslatorL4, BehavioralTranslatorL5):

  def _get_rtlir2sv_visitor( s ):
    return BehavioralRTLIRToSVVisitorL5

#-------------------------------------------------------------------------
# BehavioralRTLIRToSVVisitorL5
#-------------------------------------------------------------------------

class BehavioralRTLIRToSVVisitorL5( BehavioralRTLIRToSVVisitorL4 ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for subcomponent attributes in L5.
    """
    # Generate subcomponent attribute
    if isinstance( node.value.Type, rt.Component ) and\
       not isinstance( node.value, bir.Base ):

      value = s.visit( node.value )
      attr = node.attr
      s.check_res( node, attr )
      return f'{value}__{attr}'

    return super().visit_Attribute( node )

  #-----------------------------------------------------------------------
  # visit_Index
  #-----------------------------------------------------------------------

  def visit_Index( s, node ):
    if isinstance( node.value.Type, rt.Array ) and \
       isinstance( node.value.Type.get_sub_type(), rt.Component ):
      try:
        idx = node.idx._value
      except AttributeError:
        raise SVerilogTranslationError( s.blk, node,
          'index of component array must be a static constant expression!' )
      idx = int( idx )
      value = s.visit( node.value )
      return f"{value}__{idx}"

    else:
      return super().visit_Index( node )



#=========================================================================
# YosysBehavioralTranslatorL1.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog L1 behavioral translator."""

from pymtl3.datatypes import Bits, is_bitstruct_inst
from pymtl3.passes.backends.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.backends.sverilog.translation.behavioral.SVBehavioralTranslatorL1 import (
    BehavioralRTLIRToSVVisitorL1,
    SVBehavioralTranslatorL1,
)
from pymtl3.passes.backends.sverilog.util.utility import make_indent
from pymtl3.passes.rtlir import BehavioralRTLIR as bir
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt


class YosysBehavioralTranslatorL1( SVBehavioralTranslatorL1 ):

  def _get_rtlir2sv_visitor( s ):
    return YosysBehavioralRTLIRToSVVisitorL1

class YosysBehavioralRTLIRToSVVisitorL1( BehavioralRTLIRToSVVisitorL1 ):
  """IR visitor that generates yosys-compatible SystemVerilog code."""

  def __init__( s, is_reserved ):
    super().__init__( is_reserved )
    s.loopvars = set()

  def signal_expr_prologue( s, node ):
    # Setup the signal expression attributes
    # attr: an array of all attributes that will be filled into the attribute
    # template string `s_attr`
    # index: an array of all indices that will be filled into the index
    # template string `s_index`
    try:
      node.sexpr = node.value.sexpr
    except AttributeError:
      node.sexpr = { 'attr' : [], 'index' : [], 's_attr' : "", 's_index' : "" }

  def signal_expr_epilogue( s, node, ret ):
    try:
      # If the current node is a top level signal expression, assemble the
      # complete signal expression by filling attributes and indices into
      # their template strings
      if node._top_expr:
        all_terms = node.sexpr['attr'] + node.sexpr['index']
        template = node.sexpr['s_attr'] + node.sexpr['s_index']
        return template.format( *all_terms )
    except AttributeError:
      # If the current node is not a top level signal expression then we
      # are done. Return the given return value for debugging purposes.
      return ret

  def get_loopvars( s ):
    loopvars = list( s.loopvars )
    for i, loopvar in enumerate( loopvars ):
      loopvars[i] = "integer " + loopvar + ";"
    if loopvars:
      loopvars.append( "" )
    return loopvars

  #-----------------------------------------------------------------------
  # visit_CombUpblk
  #-----------------------------------------------------------------------

  def visit_CombUpblk( s, node ):
    upblk = super().visit_CombUpblk( node )
    return s.get_loopvars() + upblk

  #-----------------------------------------------------------------------
  # visit_SeqUpblk
  #-----------------------------------------------------------------------

  def visit_SeqUpblk( s, node ):
    upblk = super().visit_SeqUpblk( node )
    return s.get_loopvars() + upblk

  #-----------------------------------------------------------------------
  # visit_Assign
  #-----------------------------------------------------------------------

  def visit_Assign( s, node ):
    node.target._top_expr = 1
    node.value._top_expr = 1
    return super().visit_Assign( node )

  #-----------------------------------------------------------------------
  # visit_Concat
  #-----------------------------------------------------------------------

  def visit_Concat( s, node ):
    for child in node.values:
      child._top_expr = 1
    return super().visit_Concat( node )

  #-----------------------------------------------------------------------
  # visit_ZeroExt
  #-----------------------------------------------------------------------

  def visit_ZeroExt( s, node ):
    node.value._top_expr = 1
    return super().visit_ZeroExt( node )

  #-----------------------------------------------------------------------
  # visit_SignExt
  #-----------------------------------------------------------------------

  def visit_SignExt( s, node ):
    node.value._top_expr = 1
    return super().visit_SignExt( node )

  #-----------------------------------------------------------------------
  # visit_Reduce
  #-----------------------------------------------------------------------

  def visit_Reduce( s, node ):
    node.value._top_expr = 1
    return super().visit_Reduce( node )

  #-----------------------------------------------------------------------
  # visit_SizeCast
  #-----------------------------------------------------------------------

  def visit_SizeCast( s, node ):
    nbits = node.nbits
    value = None if not hasattr(node.value, "_value") else node.value._value

    if value is None:
      node.value._top_expr = 1
      value_str = s.visit( node.value )
      cur_nbits = node.value.Type.get_dtype().get_length()
      if cur_nbits == nbits:
        return value_str
      elif cur_nbits > nbits:
        msb = nbits-1
        return f"{value_str}[{msb}:0]"
      else:
        # Zero-extend the value
        n_zero = nbits - cur_nbits
        return f"{{ {{ {n_zero} {{ 1'b0 }} }}, {value_str} }}"

    if isinstance( value, Bits ):
      value = int(value)
    return f"{nbits}'d{value}"

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):

    s.signal_expr_prologue( node )

    Type = node.Type

    if isinstance(Type, rt.Const):
      obj = Type.get_object()
      if isinstance( obj, int ):
        node.sexpr['s_attr'] = f"32'd{obj}"
        node.sexpr['s_index'] = ""
      elif isinstance( obj, Bits ):
        nbits = obj.nbits
        value = int( obj )
        node.sexpr['s_attr'] = f"{nbits}'d{value}"
        node.sexpr['s_index'] = ""
      elif is_bitstruct_inst( obj ):
        node.sexpr['s_attr'] = s._struct_instance( node.Type.get_dtype(), obj )
        node.sexpr['s_index'] = ""
      else:
        raise SVerilogTranslationError( s.blk, node,
          f"{node.attr} {obj} is not an integer!" )

    elif isinstance( node.value, bir.Base ):
      # The base of this attribute node is the component 's'.
      # Example: s.out, s.STATE_IDLE
      # assert node.value.base is s.component
      s.check_res( node, node.attr )
      node.sexpr['s_attr'] = "{}"
      node.sexpr['attr'].append( node.attr )

    else:
      raise SVerilogTranslationError( s.blk, node,
          "sub-components are not supported at L1" )

    return s.signal_expr_epilogue( node, node.attr )

  #-----------------------------------------------------------------------
  # visit_Index
  #-----------------------------------------------------------------------

  def visit_Index( s, node ):

    node.idx._top_expr = 1
    idx   = s.visit( node.idx )
    value = s.visit( node.value )
    Type = node.value.Type

    s.signal_expr_prologue( node )

    # Unpacked index
    if isinstance( Type, rt.Array ):

      subtype = Type.get_sub_type()
      if isinstance( subtype, rt.Const ):
        nbits = subtype.get_dtype().get_length()
        try:
          const_value = node._value
        except AttributeError:
          raise SVerilogTranslationError( s.blk, node,
            f"{value} is not an array of constants!" )
        node.sexpr['s_index'] = f"{nbits}'d{const_value}"
        node.sexpr['index'] = []
        node.sexpr['s_attr'] = ""
        node.sexpr['attr'] = []
      else:
        node.sexpr['s_index'] += '[{}]'
        node.sexpr['index'].append( idx )

    # Index on a signal
    elif isinstance( Type, rt.Signal ):

      # Packed index or bit selection
      if Type.is_packed_indexable() or isinstance(Type.get_dtype(), rdt.Vector):
        node.sexpr['s_index'] += '[{}]'
        node.sexpr['index'].append( idx )
      else:
        raise SVerilogTranslationError( s.blk, node,
            "internal error: unrecognized index" )

    else:
      raise SVerilogTranslationError( s.blk, node,
          "internal error: unrecognized index" )

    return s.signal_expr_epilogue( node, f'{value}[{idx}]' )

  #-----------------------------------------------------------------------
  # visit_Slice
  #-----------------------------------------------------------------------

  def visit_Slice( s, node ):
    node.value._top_expr = 1
    node.lower._top_expr = 1
    node.upper._top_expr = 1
    return super().visit_Slice( node )

  #-----------------------------------------------------------------------
  # visit_FreeVar
  #-----------------------------------------------------------------------

  def visit_FreeVar( s, node ):
    if isinstance( node.obj, int ):
      return f"32'd{node.obj}"
    elif isinstance( node.obj, Bits ):
      nbits = node.obj.nbits
      value = int( node.obj )
      return f"{nbits}'d{value}"
    else:
      raise SVerilogTranslationError( s.blk, node,
      f"{node.name} {node.obj} is not an integer!" )



#=========================================================================
# SVStructuralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 4 SystemVerilog structural translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved

from ....testcases import (
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseBits32ConnectSubCompAttrComp,
    CaseConnectArraySubCompArrayStructIfcComp,
)
from ..SVStructuralTranslatorL4 import SVStructuralTranslatorL4


def run_test( case, m ):
  m.elaborate()
  SVStructuralTranslatorL4.is_sverilog_reserved = lambda s, x: x in sverilog_reserved
  tr = SVStructuralTranslatorL4( m )
  tr.clear( m )
  tr.translate_structural( m )
  subcomps = tr.structural.decl_subcomps

  check_eq( subcomps[m], case.REF_COMP )

@pytest.mark.parametrize(
  'case', [
    CaseBits32ConnectSubCompAttrComp,
    CaseConnectArraySubCompArrayStructIfcComp,
    CaseBits32ArrayConnectSubCompAttrComp,
  ]
)
def test_sverilog_structural_L4( case ):
  run_test( case, case.DUT() )



#=========================================================================
# TranslationImport_stdlib_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 3, 2019
"""Test stdlib RTL components with SystemVerilog translation and import.

We reuse all the test cases from stdlib test files. To achieve this I
overwrite the original reference to the test function used in stdlib test
files and add my own test function (the `run_test`s). However we need
to make sure the orignal reference is not lost and is restored after
finishing each test (no matter it fails or passes).
"""

from pymtl3.datatypes import Bits1, clog2, mk_bits
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.rtl.arbiters_test import test_rr_arb_4 as _rr_arb_4
from pymtl3.stdlib.rtl.arbiters_test import test_rr_arb_en_4 as _rr_arb_en_4
from pymtl3.stdlib.rtl.Crossbar_test import test_crossbar3 as _crossbar3
from pymtl3.stdlib.rtl.Encoder_test import test_encoder_5_directed as _encoder5
from pymtl3.stdlib.rtl.valrdy_queues_test import test_2entry_normal_Bits as _n2
from pymtl3.stdlib.rtl.valrdy_queues_test import test_3entry_normal_Bits as _n3
from pymtl3.stdlib.rtl.valrdy_queues_test import test_bypass_Bits as _bypass_Bits
from pymtl3.stdlib.rtl.valrdy_queues_test import test_normal_Bits as _normal_Bits
from pymtl3.stdlib.rtl.valrdy_queues_test import test_pipe_Bits as _pipe_Bits
from pymtl3.stdlib.test import TestVectorSimulator

from ..TranslationImportPass import TranslationImportPass


def local_do_test( _m ):
  try:
    _m.elaborate()
    # Mark component `_m` as to be translated and imported
    _m.sverilog_translate_import = True
    m = TranslationImportPass()( _m )
    sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
    sim.run_test()
  finally:
    try:
      # Explicitly finalize imported component to avoid shared lib name aliasing
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass

def test_arbiter_rr_arb_4( do_test ):
  # Customized test function
  def run_test( cls, args, test_vectors ):
    m = cls( *args )
    BitsN = mk_bits( args[0] )
    def tv_in( model, test_vector ):
      model.reqs = BitsN(test_vector[0])
    def tv_out( model, test_vector ):
      assert model.grants == BitsN(test_vector[1])
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  # Swap in the customized test method
  test_func = _rr_arb_4
  _run_test = test_func.__globals__['run_test']
  test_func.__globals__['run_test'] = run_test
  try:
    test_func()
  finally:
    # Restore the original method regardless of the test result
    test_func.__globals__['run_test'] = _run_test

def test_arbiter_rr_arb_en_4( do_test ):
  def run_test( cls, args, test_vectors ):
    m = cls( *args )
    BitsN = mk_bits( args[0] )
    def tv_in( model, test_vector ):
      model.en   = Bits1( test_vector[0] )
      model.reqs = BitsN( test_vector[1] )
    def tv_out( model, test_vector ):
      assert model.grants == BitsN(test_vector[2])
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _rr_arb_en_4
  _run_test = test_func.__globals__['run_en_test']
  test_func.__globals__['run_en_test'] = run_test
  try:
    test_func( None, None )
  finally:
    test_func.__globals__['run_en_test'] = _run_test

def test_crossbar3( do_test ):
  def run_test( cls, args, test_vectors ):
    m = cls( *args )
    T = args[1]
    Tsel = mk_bits( clog2( args[0] ) )

    def tv_in( model, test_vector ):
      n = len( model.in_ )
      for i in range(n):
        model.in_[i] = T(test_vector[i])
        model.sel[i] = Tsel(test_vector[n+i])
    def tv_out( model, test_vector ):
      n = len( model.in_ )
      for i in range(n):
        assert model.out[i] == T(test_vector[n*2+i])
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _crossbar3
  _run_test = test_func.__globals__['run_test_crossbar']
  test_func.__globals__['run_test_crossbar'] = run_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_crossbar'] = _run_test

def test_encoder_5_directed( do_test ):
  def run_test( m, test_vectors ):
    def tv_in( model, test_vector ):
      model.in_ = test_vector[0]
    def tv_out( model, test_vector ):
      assert model.out == test_vector[1]
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _encoder5
  _run_test = test_func.__globals__['run_test']
  test_func.__globals__['run_test'] = run_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test'] = _run_test

#-------------------------------------------------------------------------
# Queue tests
#-------------------------------------------------------------------------

def _queue_test_tv_in( model, tv ):
  model.enq.val = tv[0]
  model.enq.msg = tv[2]
  model.deq.rdy = tv[4]

def _queue_test_tv_out( model, tv ):
  if tv[1] != '?': assert model.enq.rdy == tv[1]
  if tv[3] != '?': assert model.deq.val == tv[3]
  if tv[5] != '?': assert model.deq.msg == tv[5]

def test_bypass_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _bypass_Bits
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_pipe_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _pipe_Bits
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_normal_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _normal_Bits
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_2entry_normal_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _n2
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_3entry_normal_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _n3
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test



"""
========================================================================
VcdGenerationPass.py
========================================================================

Author : Shunning Jiang, Yanghui Ou, Peitian Pan
Date   : Sep 8, 2019
"""

import os
import time
from collections import defaultdict

from pymtl3.datatypes import Bits, concat, get_nbits, to_bits
from pymtl3.dsl import Const
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError


class VcdGenerationPass( BasePass ):

  def __call__( self, top ):

    if hasattr( top, "config_tracing" ):
      top.config_tracing.check()

      if top.config_tracing.tracing != 'none':
        if not hasattr( top, "_tracing" ):
          top._tracing = PassMetadata()
        top._tracing.vcd_func = self.make_vcd_func( top, top._tracing )

  def make_vcd_func( self, top, vcdmeta ):

    vcd_file_name = top.config_tracing.vcd_file_name

    if vcd_file_name != "":
      vcdmeta.vcd_file_name = str(vcd_file_name) + ".vcd"
    else:
      vcdmeta.vcd_file_name = str(top.__class__.__name__) + ".vcd"

    vcd_file = vcdmeta.vcd_file = open( vcdmeta.vcd_file_name, "w" )

    print(f"[Tracing mode = {top.config_tracing.tracing}] "
          f"Writing value change dump (VCD) to {os.getcwd()}/{(vcdmeta.vcd_file_name)}")

    # Get vcd timescale

    try:                    vcd_timescale = top.vcd_timescale
    except AttributeError:  vcd_timescale = "10ps"

    # Print vcd header

    print( "$date\n  {}\n$end\n$version\n  PyMTL 3 (Mamba)\n$end\n"
           "$timescale\n {}\n$end\n".format( time.asctime(), vcd_timescale ),
           file=vcd_file )

    # Utility generator to create new symbols for each VCD signal.
    # Code inspired by MyHDL 0.7.
    # Shunning: I just reuse it from pymtl v2

    def _gen_vcd_symbol():

      # Generate a string containing all valid vcd symbol characters
      _codechars = ''.join([chr(i) for i in range(33, 127)])
      _mod       = len(_codechars)

      # Generator logic
      n = 0
      while True:
        q, r = divmod(n, _mod)
        code = _codechars[r]
        while q > 0:
          q, r = divmod(q, _mod)
          code = _codechars[r] + code
        yield code
        n += 1

    vcd_symbols = _gen_vcd_symbol()

    # Preprocess some metadata

    component_signals = defaultdict(set)

    all_components = set()

    # We only collect top level signals, and squash bitstruct into a long
    # bits object
    for x in top._dsl.all_signals:
      if x.is_top_level_signal():
        host = x.get_host_component()
        component_signals[ host ].add( x )

    # We pre-process all nets in order to remove all sliced wires because
    # they belong to a top level wire and we count that wire

    trimmed_value_nets = []
    vcdmeta.vcd_clock_net_idx = None

    # FIXME handle the case where the top level signal is in a value net
    for writer, net in top.get_all_value_nets():
      new_net = []
      for x in net:
        if not isinstance(x, Const) and x.is_top_level_signal():
          new_net.append( x )
          if repr(x) == "s.clk":
            # Hardcode clock net because it needs to go up and down
            assert vcdmeta.vcd_clock_net_idx is None
            vcdmeta.vcd_clock_net_idx = len(trimmed_value_nets)

      if new_net:
        trimmed_value_nets.append( new_net )

    # Generate symbol for existing nets

    net_symbol_mapping = [ next(vcd_symbols) for x in trimmed_value_nets ]
    signal_net_mapping = {}

    for i in range(len(trimmed_value_nets)):
      for x in trimmed_value_nets[i]:
        signal_net_mapping[x] = i

    # Inner utility function to perform recursive descent of the model.
    # Shunning: I mostly follow v2's implementation

    # Vcd file takes a(0) instead of a[0]
    def vcd_mangle_name( name ):
      # signal names with colons in it silently fail gtkwave
      return name.replace('[','(').replace(']',')').replace(':', '__')

    def recurse_models( m, spaces ):

      # Special case the top level "s" to "top"

      my_name = m.get_field_name()
      if my_name == "s":
        my_name = "top"

      # Create a new scope for this module
      print( f"{spaces}$scope module {vcd_mangle_name(my_name)} $end",
             file=vcd_file )

      m_name = repr(m)

      # Define all signals for this model.
      for signal in component_signals[m]:

        # Multiple signals may be collapsed into a single net in the
        # simulator if they are connected. Generate new vcd symbols per
        # net, not per signal as an optimization.

        if signal in signal_net_mapping:
          net_id = signal_net_mapping[signal]
          symbol = net_symbol_mapping[net_id]
        else:
          # We treat this as a new net

          # Check if it's clock. Hardcode clock net
          if repr(signal) == "s.clk":
            assert vcdmeta.vcd_clock_net_idx is None
            vcdmeta.vcd_clock_net_idx = len(trimmed_value_nets)

          # This is a signal whose connection is not captured by the
          # global net data structure. This might be a sliced signal or
          # a signal updated in an upblk. Creating a new net for it does
          # not hurt functionality.

          trimmed_value_nets.append( [ signal ] )
          signal_net_mapping[signal] = len(signal_net_mapping)
          symbol = next(vcd_symbols)
          net_symbol_mapping.append( symbol )

        # This signal can be a part of an interface so we have to
        # "subtract" host component's name from signal's full name
        # to get the actual name like enq.rdy
        # TODO struct
        signal_name = vcd_mangle_name( repr(signal)[ len(m_name)+1: ] )
        print( f"{spaces}  $var reg {get_nbits(signal._dsl.Type)} {symbol} {signal_name} $end",
               file=vcd_file )

      # Recursively visit all submodels.
      for child in m.get_child_components():
        recurse_models( child, spaces+'  ' )

      print( f"{spaces}$upscope $end", file=vcd_file )

    # Begin recursive descent from the top-level model.
    recurse_models( top, '' )

    # Once all models and their signals have been defined, end the
    # definition section of the vcd and print the initial values of all
    # nets in the design.
    print( "$enddefinitions $end\n", file=vcd_file )

    # vcdmeta.last_values is an array of values from the previous cycle

    last_values = vcdmeta.last_values = [0 for _ in range(len(trimmed_value_nets))]

    for i, net in enumerate(trimmed_value_nets):
      # Convert everything to Bits to get around lack of bit struct support.
      # The first cycle VCD contains the default value
      bin_str = to_bits( net[0]._dsl.Type() ).bin()

      print( f"b{bin_str} {net_symbol_mapping[i]}", file=vcd_file )

      # Set this to be the last cycle value str
      last_values[i] = bin_str

    # Now we create per-cycle signal value collect functions

    vcdmeta.vcd_sim_ncycles = 0

    # Separate clock net from normal nets ahead of time
    clock_symbol = net_symbol_mapping[ vcdmeta.vcd_clock_net_idx ]

    net_details = [ ( trimmed_value_nets[i][0], net_symbol_mapping[i] )
                    for i in range(len(trimmed_value_nets))
                      if i != vcdmeta.vcd_clock_net_idx ]

    # Flip clock for the first cycle
    print( '\n#0\nb0b1 {}\n'.format( clock_symbol ), file=vcd_file, flush=True )

    # Returns a dump_vcd function that is ready to be appended to _sched.
    # TODO: type check?

    # Adding this 's' argument is for eval to correctly evaluate 's.x'...
    # Python 3 destroys a lot of our hacks .. sigh

    def dump_vcd_inner( s ):

      for i, (signal, symbol) in enumerate( net_details ):

        # If we encounter a BitStruct then dump it as a concatenation of
        # all fields.
        # TODO: treat each field in a BitStruct as a separate signal?

        try:
          net_bits_bin = to_bits( eval(repr(signal)) )
        except Exception as e:
          raise TypeError(f'{e}\n - {signal} becomes another type. Please check your code.')

        net_bits_bin_str = net_bits_bin.bin()
        # `last_value` is the string form of a Bits object in binary
        # e.g. '0b000' == Bits3(0).bin()
        # We store strings instead of values ...
        if last_values[i] != net_bits_bin_str:
          last_values[i] = net_bits_bin_str
          print( f'b{net_bits_bin_str} {symbol}', file=vcd_file )

      # Flop clock at the end of cycle
      next_neg_edge = 100 * vcdmeta.vcd_sim_ncycles + 50
      print( f'\n#{next_neg_edge}\nb0b0 {clock_symbol}', file=vcd_file )

      # Flip clock of the next cycle
      next_pos_edge = next_neg_edge + 50
      print( f'#{next_pos_edge}\nb0b1 {clock_symbol}\n', file=vcd_file, flush=True )
      vcdmeta.vcd_sim_ncycles += 1

    def gen_dump_vcd( s ):
      def dump_vcd():
        dump_vcd_inner( s )
      return dump_vcd

    return gen_dump_vcd( top )



"""
==========================================================================
bitstruct_test.py
==========================================================================
Test cases for bitstructs.

Author : Yanghui Ou
  Date : July 27, 2019
"""
from copy import deepcopy

import pytest

from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.dsl.test.sim_utils import simple_sim_pass

from ..bits_import import *
from ..bitstructs import bitstruct, is_bitstruct_class, is_bitstruct_inst, mk_bitstruct

#-------------------------------------------------------------------------
# Basic test to test error messages and exceptions
#-------------------------------------------------------------------------

def test_no_field():
  try:
    @bitstruct
    class Pixel:
      b = Bits8
  except AttributeError as e:
    print(e)
    assert str(e).startswith( "No field is declared in the bit struct definition." )

def test_field_no_default():
  try:
    @bitstruct
    class Pixel:
      r : Bits8
      g : Bits8
      b : Bits8 = b8(4)
  except TypeError as e:
    print(e)
    assert str(e).startswith( "We don't allow subfields to have default value:\n- Field " )

def test_field_wrong_type():
  try:
    @bitstruct
    class A:
      x: int
  except TypeError as e:
    print(e)
    assert str(e).startswith( "We currently only support BitsN, list, or another BitStruct as BitStruct field:\n- Field " )

def test_field_not_type():
  try:
    @bitstruct
    class A:
      x: 1
  except TypeError as e:
    print(e)

@bitstruct
class Pixel:
  r : Bits8
  g : Bits8
  b : Bits8
  nbits = 24

MadePixel = mk_bitstruct( 'MadePixel',{
    'r' : Bits8,
    'g' : Bits8,
    'b' : Bits8,
  },
  namespace = { 'nbits' : 24 }
)

#-------------------------------------------------------------------------
# Caching test
#-------------------------------------------------------------------------

def test_structs_caching():

  class A:
    @bitstruct
    class S:
      x: Bits8
      y: Bits8
      z:  [ [ Bits32, Bits32 ] ] * 2

  class B:
    @bitstruct
    class S:
      x: Bits8
      y: Bits8
      z:  [ [ Bits32, Bits32 ] ] * 2

  SS = mk_bitstruct( 'S', {
    'x' : Bits8,
    'y' : Bits8,
    'z' : [ [ Bits32, Bits32 ], [ Bits32, Bits32 ] ]
  })

  assert B.S is A.S
  assert SS is B.S
  assert SS is A.S

  class C:
    @bitstruct
    class S:
      x: Bits8
      y: Bits8
      w:  [ [ Bits32, Bits32 ] ] * 2

  assert C.S is not A.S

  SS2 = mk_bitstruct( 'S', {
    'x' : Bits8,
    'y' : Bits7,
    'z' : [ [ Bits32, Bits32 ], [ Bits32, Bits32 ] ]
  })

  assert SS2 is not A.S

  SS3 = mk_bitstruct( 's', {
    'x' : Bits8,
    'y' : Bits8,
    'z' : [ [ Bits32, Bits32 ], [ Bits32, Bits32 ] ]
  })
  assert SS3 is not A.S

# Shunning: As of now, since we still want metadata fields in the struct
# definition, we have to mark this as xfail until we find a better
# solution. In terms of the test, B.S gets cached because A.S and B.S have
# the same annotated fields. However, the nbits field of B is different
# from A, which is an underfined behavior ...

@pytest.mark.xfail
def test_structs_caching_metadata_undefined():

  class A:
    @bitstruct
    class S:
      x: Bits8
      y: Bits8
      nbits = 1

  class B:
    @bitstruct
    class S:
      x: Bits8
      y: Bits8
      nbits = 2

  assert B.S is A.S
  assert B.S.nbits == 2 # From get cache B.S is A.S ... so no nbits=2

#-------------------------------------------------------------------------
# Test the functionalities
#-------------------------------------------------------------------------

def test_simple():
  print()

  # Test basic
  px = Pixel()
  assert isinstance( px.r, Bits8 )
  assert isinstance( px.g, Bits8 )
  assert isinstance( px.b, Bits8 )
  assert px.r == px.g == 0
  assert px.b == 0
  assert px.nbits == 24

  # Test dynamic basic
  mpx = MadePixel()
  assert isinstance( mpx.r, Bits8 )
  assert isinstance( mpx.g, Bits8 )
  assert isinstance( mpx.b, Bits8 )
  assert mpx.r == mpx.g == 0
  assert mpx.b == 0
  assert mpx.nbits == 24

  # Test str
  assert str(px) == str(mpx)

  # Check repr
  print(( repr(px), repr(mpx) ))

  # test equality
  px1 = Pixel(1, 2, 3)
  assert isinstance( px1.r, Bits8 )
  assert isinstance( px1.g, Bits8 )
  assert isinstance( px1.b, Bits8 )
  px2 = Pixel( b4(0), b4(0), b4(0) )
  assert px != px1
  assert px == px2

# FIXME: The following inherited bit struct cannot be used for construct
# nested struct as the newly made class is not captured in the scope of
# mk_bitstruct. Manually created structs cannot be used by mk_bitstruct
# to create nested structs.
# class StaticPoint(
#   mk_bitstruct( "BasePoint", [
#     ( 'x', Bits4 ),
#     ( 'y', Bits4 ),
#   ]) ):
#
#   def __str__( s ):
#     return "({},{})".format( int(s.x), int(s.y) )

# Create bit struct using syntax sugar (mk_bitstruct):
StaticPoint = mk_bitstruct( "StaticPoint", {
    'x': Bits4,
    'y': Bits4,
  })
# Create nested struct using mk_bitstruct
NestedSimple = mk_bitstruct( "NestedSimple", {
    'pt0': StaticPoint,
    'pt1': StaticPoint,
  })

def test_struct():
  pt = StaticPoint( Bits4(2), Bits4(4) )
  print( pt           )
  # print( pt.to_bits() )
  try:
    StaticPoint.__dict__[ 'haha' ]
  except KeyError as e:
    assert str( e ) == "'haha'"
  # assert pt.to_bits() == 0x24
  assert pt.x == 2
  assert pt.y == 4

  np = NestedSimple( StaticPoint(1,2), StaticPoint(3,4) )
  print( NestedSimple )
  print( np           )
  assert np.pt0 == StaticPoint(1,2)
  assert np.pt0 != StaticPoint(1,1)
  assert np.pt0.x == 1
  assert np.pt0.y == 2
  assert np.pt1.x == 3
  assert np.pt1.y == 4

  def dynamic_point( xnbits, ynbits ):
    XType = mk_bits( xnbits )
    YType = mk_bits( ynbits )

    return mk_bitstruct(
      f"Point_{xnbits}_{ynbits}", {
        'x': XType,
        'y': YType,
      },
      namespace = { '__str__': lambda s: "({},{})".format( int(s.x), int(s.y) ) }
    )

  DPoint = dynamic_point( 4, 8 )
  dp = DPoint( 1, 2 )
  print( DPoint )
  print( dp     )
  assert dp.x == 1
  assert dp.y == 2

def test_component():
  class A( Component ):
    def construct( s ):
      s.in_    =  InPort( NestedSimple )
      s.out    = OutPort( NestedSimple )
      s.out_pt = OutPort( StaticPoint  )

      @s.update
      def up_bitstruct():
        # TODO:We have to use deepcopy here as a temporary workaround
        # to prevent s.in_ being mutated by the following operations.
        s.out = deepcopy( s.in_ )
        s.out.pt0.x += 1
        s.out.pt0.y -= 1
        s.out.pt1.x += 1
        s.out.pt1.y -= 1
        s.out_pt = s.in_.pt1

  dut = A()
  dut.elaborate()
  dut.apply( simple_sim_pass )
  dut.in_ = NestedSimple( StaticPoint(b4(1),b4(2)), StaticPoint(b4(3),b4(4)) )
  dut.tick()
  assert dut.out == NestedSimple( StaticPoint(Bits4(2),Bits4(1)), StaticPoint(Bits4(4),Bits4(3)) )
  print( dut.out_pt )
  assert dut.out_pt == StaticPoint(Bits4(3),Bits4(4))

#-------------------------------------------------------------------------
# Overwrite test
#-------------------------------------------------------------------------

@bitstruct
class Point:
  x : Bits8
  y : Bits8

  def __str__( self ):
    return f'({int(self.x)},{int(self.y)})'

  def __eq__( self, other ):
    return self.x == other.x and self.y == other.y

MadePoint = mk_bitstruct( 'MadePoint', {
    'x' : Bits8,
    'y' : Bits8,
  },
  namespace = {
    '__str__' : lambda self : f'({int(self.x)},{int(self.y)})',
    '__eq__'  : lambda self, other : self.x == other.x and self.y == other.y
  })

def test_overwrite():
  pt  = Point( b8(1), b8(2) )
  mpt = MadePoint( b8(1), b8(2) )

  assert pt == mpt
  assert str(pt) == str(mpt)

#-------------------------------------------------------------------------
# Nested bit struct test
#-------------------------------------------------------------------------

@bitstruct
class TwoPoint:
  pt0 : Point
  pt1 : MadePoint

MadeTwoPoint = mk_bitstruct( 'MadeTwoPoint', {
  'pt0' : Point,
  'pt1' : MadePoint,
})

def test_nested():
  a = TwoPoint( pt0=Point( b8(1), b8(2) ) )
  b = MadeTwoPoint( pt1=MadePoint( b8(3), b8(4) ) )
  assert a.pt0.x == 1
  assert a.pt0.y == 2
  assert a.pt1.x == 0
  assert a.pt1.y == 0
  assert b.pt0.x == 0
  assert b.pt0.y == 0
  assert b.pt1.x == 3
  assert b.pt1.y == 4

def test_nested_two_struct_with_same_name():

  class A:
    @bitstruct
    class S:
      z: Bits2

  class B:
    @bitstruct
    class S:
      x: Bits8
      y: Bits8

  SS = mk_bitstruct( 'SS', {
    'a' : A.S,
    'b' : B.S,
  })
  ss = SS()
  assert isinstance( ss.a.z, Bits2 )
  assert isinstance( ss.b.x, Bits8 )
  assert isinstance( ss.b.y, Bits8 )

  ss = SS( A.S(2), B.S(3,3) )
  assert isinstance( ss.a.z, Bits2 )
  assert isinstance( ss.b.x, Bits8 )
  assert isinstance( ss.b.y, Bits8 )

  assert ss.a.z == 2
  assert ss.b.x == 3
  assert ss.b.y == 3

#-------------------------------------------------------------------------
# Bit struct instance test
#-------------------------------------------------------------------------

def test_is_bitstruct_inst_class():
  class A:
    x : Bits4
    y : Bits4

  @bitstruct
  class B:
    x : Bits4
    y : Bits4

  assert not is_bitstruct_class( A )
  assert not is_bitstruct_inst( A )
  assert not is_bitstruct_class( A() )
  assert not is_bitstruct_inst( A() )

  assert is_bitstruct_class( B )
  assert not is_bitstruct_inst( B )
  assert not is_bitstruct_class( B() )
  assert is_bitstruct_inst( B() )

#-------------------------------------------------------------------------
# bit struct with array test
#-------------------------------------------------------------------------

def test_list_same_class():
  @bitstruct
  class A:
    x: Bits4
    y: [ Bits4, Bits4 ]
  a = A()
  assert a.x == Bits4(0)
  assert a.y == [ Bits4(0), Bits4(0) ]

  b = a.clone()
  assert a == b

  b.x = Bits4(3)

  assert a.x != b.x
  assert a.y[0] == b.y[0]
  assert a.y[1] == b.y[1]
  assert a != b

  b.y[0] = Bits4(4)

  assert a.x != b.x
  assert a.y[0] != b.y[0]
  assert a.y[1] == b.y[1]
  assert a != b

  b.y[1] = Bits4(4)

  assert a.x != b.x
  assert a.y[0] != b.y[0]
  assert a.y[1] != b.y[1]
  assert a != b

def test_crazy_list_not_same_class():
  @bitstruct
  class A:
    x: Bits4
  try:
    @bitstruct
    class B:
      x: Bits4
      y: [[[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, Bits1]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]],
          [[[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]], [[A, A, A]]]]
  except TypeError as e:
    print(e)
    assert str(e) == "The provided list spec should be a strict multidimensional ARRAY with no varying sizes or types. "\
                     "All non-list elements should be VALID types."

def test_high_d_list_inside():
  @bitstruct
  class A:
    x: Bits4
    y:[ [ [ [Bits4, Bits4, Bits4] ] ] * 6 ] * 10
  a = A()
  assert a.x == Bits4(0)
  assert a.y == [ [ [ [ Bits4(), Bits4(), Bits4()] ] for _ in range(6) ] for _ in range(10) ]

def test_mk_high_d_list_inside():
  A = mk_bitstruct( "A", {
    'x': Bits4,
    'y': [ [ [ [Bits4, Bits4, Bits4] ] ] * 6 ] * 10,
  })
  a = A()
  assert a.x == Bits4(0)
  assert a.y == [ [ [ [ Bits4(), Bits4(), Bits4()] ] for _ in range(6) ] for _ in range(10) ]

  print(a)

def test_high_d_list_struct_inside():
  @bitstruct
  class A:
    x: Bits4
  @bitstruct
  class B:
    x: Bits4
    y: [ [ [ [A, A, A] ] ] * 6 ] * 10
  b = B()
  assert b.x == Bits4(0)
  assert b.y == [ [ [ [ A(), A(), A()] ] for _ in range(6) ] for _ in range(10) ]

def test_mk_high_d_list_struct_inside():
  @bitstruct
  class A:
    x: Bits4
  B = mk_bitstruct( "B", {
    'x': Bits4,
    'y': [ [ [ [A, A, A] ] ] * 6 ] * 10,
  })
  b = B()
  assert b.x == Bits4(0)
  assert b.y == [ [ [ [ A(), A(), A()] ] for _ in range(6) ] for _ in range(10) ]

  print(b)
  c = b.clone()
  c.y[9][2][0][2].x = Bits4(3)
  print(c.y[9][2][0][2])
  assert b != c
